import Char
import Monad 

--------------------------------------------------------------------
-- Data Types
--------------------------------------------------------------------

data XToken = Id String | AssOp | Num Integer |
             LeftParenth | RightParenth |
             Plus | Minus | Mult | Div |
             Equal | Unequal | Greater | Less |
             BeginSymb | EndSymb |
             IfSymb | ThenSymb | ElseSymb | FiSymb |
             WhileSymb | DoSymb | OdSymb |
             SemicolonSymb |
             Err
             deriving (Show)
	     
-- Monadic Parser
newtype Parser a = Parser (String -> [(a,String)])

-- Combinator Parser
type ParserPC a b = [a] -> [([a],b)]


--------------------------------------------------------------------
-- Monadic Parser
--------------------------------------------------------------------

-- Main Parsing Function
main_xpm :: String -> [XToken]
main_xpm input
	| length result == 0 	= [Err]
	| otherwise			=  fst (head result)
	where
	result = filter (\ (token_pms,rest) -> rest == "") $ (apply prog_pm) input

prog_pm = do {symb "begin"; s <- stmt_pm; symb "end";
	return ([BeginSymb]++s++[EndSymb])} 

stmt_pm = assstmt_pm +++ ifstmt_pm +++ whilestmt_pm +++ compstmt_pm

assstmt_pm = do {id <- idf_pm; symb ":="; ex <- aExpr_pm; 
	return (id++[AssOp]++ex)}

ifstmt_pm = do {symb "if"; b <- bExpr_pm; symb "then"; s1 <- stmt_pm; symb "else"; s2 <- stmt_pm; 
	symb "fi"; return ([IfSymb] ++ b ++ [ThenSymb] ++ s1 ++ [ElseSymb] ++ s2 ++ [FiSymb])}
	
whilestmt_pm = do { symb "while"; b <- bExpr_pm; symb "do"; s <- stmt_pm; symb "od"; 
	return ([WhileSymb]++b++[DoSymb]++s++[OdSymb])}

compstmt_pm = do { symb "("; s1 <- stmt_pm; symb ";"; s2 <- stmt_pm; symb ")"; 
	return ([LeftParenth]++s1++[SemicolonSymb]++s2++[RightParenth])}
	
expr_pm = aExpr_pm +++ bExpr_pm

aExpr_pm = do
	t <- term_pm; 
	tf <- many_pm(do{op_pm <- aOp_pm; t <- term_pm; return (op_pm++t)}); 
	return (t++(concat tf))

term_pm = do
	f <- factor_pm; 
	mf <- many_pm(do{op_pm <- mOp_pm; f <- factor_pm; return (op_pm++f)}); 
	return (f++(concat mf))

factor_pm = op_pm +++ do{symb "("; ex <- aExpr_pm; symb ")"; 
	return ([LeftParenth]++ex++[RightParenth])}

op_pm = idf_pm +++ numeral_pm

aOp_pm = do {symb "+"; return [Plus]} +++ do {symb "-"; return [Minus]} 

mOp_pm = do {symb "*"; return [Mult]} +++ do {symb "/"; return [Div]} 

bExpr_pm = do { symb "("; ex1 <- aExpr_pm; r <- relOp_pm; ex2 <- aExpr_pm; symb ")"; 
	return ([LeftParenth]++ex1++r++ex2++[RightParenth])} 

relOp_pm = do {symb "="; return [Equal]} +++ do {symb "<"; return [Less]} +++ 
	do {symb ">"; return [Greater]} +++ do {symb "/="; return [Unequal]} 

idf_pm = do {c <- (sat isAlpha); cs <- many_pm (sat isAlphaNum); space; return ([Id (c:cs)])}

numeral_pm = do{cs <- many1_pm(sat isDigit); space; return [Num (read cs)]}

--------------------------------------------------------------------
-- Combinator Parser
--------------------------------------------------------------------

-- Main Parsing Function
main_xpc :: String -> [XToken]
main_xpc input
	| length result == 0 	= [Err]
	| otherwise			=  snd (head result)
	where
	result = filter (\ (rest,tokens) -> rest == "") $ prog input
	

prog = wsp (token "begin" <*> stmt <*> token "end") <@ makeProg
makeProg (b,(st,e)) = [BeginSymb] ++ st ++ [EndSymb]

stmt = assStmt <|> ifStmt <|> whileStmt <|> compStmt

assStmt = (idf <*> token ":=" <*> aExpr) <@ makeAssStmt
makeAssStmt (idf, (assOp, aExpr)) = idf ++ [AssOp] ++ aExpr

ifStmt = token "if" <*> bExpr <*> token "then" <*> stmt <*> token "else" <*> stmt <*> token "fi" <@ makeIf
makeIf (i,(e,(t,(s1,(el,(s2,fi)))))) = [IfSymb] ++ e ++ [ThenSymb] ++ s1 ++ [ElseSymb] ++ s2 ++ [FiSymb]

whileStmt = token "while" <*> bExpr <*> token "do" <*> stmt <*> token "od" <@ makeWhile
makeWhile (w,(e,(d,(s,od)))) = [WhileSymb] ++ e ++ [DoSymb] ++ s ++ [OdSymb]

compStmt = symbol '(' <*> stmt <*> symbol ';' <*> stmt <*> symbol ')' <@ makeCompStmt
makeCompStmt (l,(s1,(k,(s2,r)))) = [LeftParenth] ++ s1 ++ [SemicolonSymb] ++ s2 ++ [RightParenth]

expr = aExpr <|> bExpr

aExpr = term2 <*> (many aopTerm <@ concat) <@ dePairLists
aopTerm = aop <*> term2 <@ dePairLists

term2 = factor <*> (many mopFactor <@ concat) <@ dePairLists
mopFactor = mop <*> factor <@ dePairLists

dePairLists (a,b) = (a++b)


factor = opd <|>  (symbol '(' <*> aExpr <*> symbol ')') <@ makeFactor
makeFactor (leftP,(e,rightP)) = [LeftParenth] ++ e ++ [RightParenth]

opd = numeral <|> idf

aop = (symbol '+' <|> symbol '-') <@ makeAop
makeAop c
      | c == '+'	= [Plus]
      | c == '-'	= [Minus]

mop = (symbol '*' <|> symbol '/') <@ makeMop
makeMop c
      | c == '*'	= [Mult]
      | c == '/'	= [Div]
      
      
bExpr = (symbol '(' <*> aExpr <*> relop <*> aExpr <*> symbol ')') <@ makebExpr
makebExpr (l,(a1,(ro,(a2,r)))) = [LeftParenth] ++ a1 ++ ro ++ a2 ++ [RightParenth]

relop = (token "=" <|> token "/=" <|> token ">" <|> token "<") <@ makeRelOp
makeRelOp s 
      | s == "="  = [Equal]
      | s == "<"  = [Less]
      | s == ">"  = [Greater]
      | s == "/=" = [Unequal]
      
idf = (satisfy isAlpha <*> (many (satisfy isAlphaNum))) <@ (\(x,xs) -> [Id (x:xs)])

digit :: ParserPC Char Integer
digit = satisfy isDigit <@ toInteger.digitToInt

numeral = many1 (satisfy isDigit) <@ (\x -> [Num (read x)])


--------------------------------------------------------------------
-- Monadic Parser Lib
--------------------------------------------------------------------

-- deconstructor
parse (Parser p) = p

instance Monad Parser where
	-- The always successful parser
	return a = Parser (\cs -> [(a,cs)])
	-- Sequences
	p >>= f = Parser (\cs -> concat [parse (f a) cs'|
		(a,cs') <- parse p cs])
		
instance MonadPlus Parser where
	mplus p q = Parser (\cs -> parse p cs ++ parse q cs)
	mzero = Parser (\cs -> [])
	
infixr 5 +++
(+++) :: Parser a -> Parser a -> Parser a
p +++ q = Parser (\cs -> case parse (mplus p q) cs of
	[] -> []
	(x:xs) -> [x])
	
item :: Parser Char
item = Parser (\cs -> case cs of
	"" -> []
	(c:cs) -> [(c,cs)])
	
sat :: (Char -> Bool) -> Parser Char
sat p = do {c <- item; if p c then return c else mzero}

force           :: Parser a -> Parser a
force p          = Parser (\cs -> let xs = parse p cs in
                              (fst (head xs), snd (head xs)) : tail xs)
	
char :: Char -> Parser Char
char c = sat (c ==)

string          :: String -> Parser String
string ""        = return ""
string (c:cs)    = do {char c; string cs; return (c:cs)}

many_pm            :: Parser a -> Parser [a]
many_pm p           = force (many1_pm p +++ return [])

many1_pm           :: Parser a -> Parser [a]
many1_pm p          = do {a <- p; as <- many_pm p; return (a:as)}

space           :: Parser String
space            = many_pm (sat isSpace)

token_pm           :: Parser a -> Parser a
token_pm p          = do {a <- p; space; return a}

symb            :: String -> Parser String
symb cs          = token_pm (string cs)

apply :: Parser a -> String -> [(a,String)]
apply p = parse (do {space; p})

--------------------------------------------------------------------
-- Combinator Parser Lib
--------------------------------------------------------------------

infixr 6 <*>
infixl 5 <@ 
infixr 4 <|>

-- char identisch
symbol :: Eq s => s -> ParserPC s s
symbol a []                      = []
symbol a (x:xs)  | a ==x         = [(xs,x)]
                 | otherwise     = []

-- string identisch
token :: Eq s => [s] -> ParserPC s [s]
token k xs | k == (take n xs)    = [(drop n xs, k)] 
           | otherwise           = []
                                 where n = length k

-- praedikat erfuellt
satisfy :: (s -> Bool) -> ParserPC s s
satisfy p [] = []
satisfy p (x:xs) = [(xs,x) | p x]

-- eingabe wird nicht verwendet
epsilon :: ParserPC s ()
epsilon xs = [(xs,())]
 
succeed :: r -> ParserPC s r
succeed v xs = [(xs,v)]

pfail :: ParserPC s r
pfail xs = []

(<*>) :: ParserPC s a -> ParserPC s b -> ParserPC s (a,b)
(p1 <*> p2) xs = [(xs2, (v1,v2)) 
                 | (xs1,v1) <- p1 xs,
                   (xs2,v2) <- p2 xs1]
	   

(<|>) :: ParserPC s a -> ParserPC s a -> ParserPC s a
(p1 <|> p2) xs = p1 xs ++ p2 xs

--  wsp entfernt Leerzeichen  
wsp :: ParserPC Char a -> ParserPC Char a
wsp p = p . filter (\x -> not (x == ' ' || x == '\n'))

(<@) :: ParserPC s a -> (a-> b) -> ParserPC s b
(p <@ f) xs = [(ys, f v) | (ys,v) <- p xs]

list (x,xs) = x:xs

many :: ParserPC s a -> ParserPC s [a]
many p = p <*> 
	many p <@ list <|> succeed []

many1 p = p <*> many p <@ list
